#include <zephyr/types.h>
#include <zephyr/kernel.h>

#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <soc.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>

#include <drivers/counter.h>
#include <drivers/gpio.h>

#include <stdio.h>

#include <zephyr/logging/log.h>

#define LOG_MODULE_NAME fast_notifications
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

#define DEVICE_NAME CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN (sizeof(DEVICE_NAME) - 1)

#define BT_UUID_CUST_SERVICE_VAL \
	BT_UUID_128_ENCODE(0xAA962F00, 0x3E10, 0x4225, 0x90A4, 0x1E40E9E27895)
#define BT_UUID_CUST_CHAR_VAL \
	BT_UUID_128_ENCODE(0xAA962F00, 0x3E10, 0x4225, 0x90A4, 0x1E40E9E27895)

#define BT_UUID_CUST_SERVICE BT_UUID_DECLARE_128(BT_UUID_CUST_SERVICE_VAL)
#define BT_UUID_CUST_CHAR BT_UUID_DECLARE_128(BT_UUID_CUST_CHAR_VAL)

static struct bt_conn *conn_ = NULL;

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_CUST_SERVICE_VAL),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};

static const struct bt_data sd[] = {
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};

void ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value)
{
    LOG_DBG("Notification has been turned %s", value == BT_GATT_CCC_NOTIFY ? "on" : "off");
}

BT_GATT_SERVICE_DEFINE(cust_svc, BT_GATT_PRIMARY_SERVICE(BT_UUID_CUST_SERVICE),
                       BT_GATT_CHARACTERISTIC(BT_UUID_CUST_CHAR, BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_READ, NULL, NULL, NULL),
                       BT_GATT_CCC(ccc_cfg_changed,
                                   BT_GATT_PERM_READ | BT_GATT_PERM_WRITE));

static void ble_mtu_exchanged(struct bt_conn *conn, uint8_t err,
								 struct bt_gatt_exchange_params *params)
{
	if (!err)
	{
		LOG_INF("MTU exchange done (%d)", bt_gatt_get_mtu(conn));
	}
	else
	{
		LOG_WRN("MTU exchange failed (err %d)", err);
	}
}

void ble_connected(struct bt_conn *conn, uint8_t hci_err)
{
	struct bt_conn_info info = {0};
	int err;

	if (hci_err)
	{
		if (hci_err == BT_HCI_ERR_UNKNOWN_CONN_ID) return;
		LOG_ERR("Connection failed (err 0x%02x)", hci_err);
		return;
	}

	if (conn_)
	{
		LOG_WRN("Connection exists, disconnect second connection\n");
		bt_conn_disconnect(conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
		return;
	}

	conn_ = bt_conn_ref(conn);

	err = bt_conn_get_info(conn_, &info);
	if (err)
	{
		LOG_ERR("Failed to get connection info %d", err);
		return;
	}

	LOG_INF("Connected (interval: %d)", info.le.interval);

    static struct bt_gatt_exchange_params exchange_params;
    exchange_params.func = ble_mtu_exchanged;
	err = bt_gatt_exchange_mtu(conn_, &exchange_params);
    if (err)
    {
        LOG_WRN("MTU exchange failed (err %d)", err);
    }
    else
    {
        LOG_INF("MTU exchange pending");
    }
}

void ble_le_param_updated(struct bt_conn *conn, uint16_t interval, uint16_t latency,
							 uint16_t timeout)
{
	LOG_INF("Connection parameters updated.");
	LOG_INF("Interval: %d, latency: %d, timeout: %d", interval, latency, timeout);
}

void ble_le_data_length_updated(struct bt_conn *conn, struct bt_conn_le_data_len_info *info)
{
	LOG_INF("LE data len updated: TX (len: %d time: %d) RX (len: %d time: %d)", info->tx_max_len,
			info->tx_max_time, info->rx_max_len, info->rx_max_time);
}

BT_CONN_CB_DEFINE(conn_callbacks) = {.connected = ble_connected,
									 .le_param_updated = ble_le_param_updated,
									 .le_data_len_updated = ble_le_data_length_updated};

struct bt_conn *ble_conn(void)
{
	return conn_;
}

int ble_adv_start(void)
{
	struct bt_le_adv_param *adv_param = BT_LE_ADV_PARAM(
		BT_LE_ADV_OPT_CONNECTABLE, BT_GAP_ADV_FAST_INT_MIN_2, BT_GAP_ADV_FAST_INT_MAX_2, NULL);

	return bt_le_adv_start(adv_param, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
}

static K_SEM_DEFINE(ready_, 0, 1);

static uint8_t index_ = 0;
static uint8_t p_in_[240] = "  Hello World";
static uint8_t p_out_[240];

static void counter_interrupt_fn(const struct device *counter_dev, void *user_data)
{
	p_in_[0] = index_++;
    k_sem_give(&ready_);
}

#define GPIO_NODE DT_NODELABEL(gpio0)
static const struct device *gpio_ = DEVICE_DT_GET(GPIO_NODE);

void gpio_toggle(int pin)
{
    gpio_pin_toggle(gpio_, pin);
}

#define TIMER_NODE DT_NODELABEL(timer1)
static const struct device *timer_ = DEVICE_DT_GET(TIMER_NODE);

void main(void)
{
	bt_enable(NULL);
	ble_adv_start();

	gpio_pin_configure(gpio_, 4, GPIO_OUTPUT);

	const struct counter_top_cfg timer_cfg = {
        .ticks = counter_us_to_ticks(timer_, 5000),
        .callback = counter_interrupt_fn,
    };

    counter_set_top_value(timer_, &timer_cfg);
    counter_start(timer_);

	for (;;)
	{
		// Take semaphore
        k_sem_take(&ready_, K_FOREVER);

		// Copy data
		memcpy(p_out_, p_in_, sizeof(p_out_));

		// Notify
		struct bt_gatt_notify_params params = {0};
		const struct bt_gatt_attr *attr = &cust_svc.attrs[2];

		params.attr = attr;
		params.data = p_out_;
		params.len = sizeof(p_out_);

		if (bt_gatt_is_subscribed(conn_, attr, BT_GATT_CCC_NOTIFY))
		{
			gpio_toggle(4);
			bt_gatt_notify_cb(conn_, &params);
			gpio_toggle(4);
		}
	}
}